"""This module communicates with SigmaStudio.

It can receive read/write requests and return with read response packets.
"""
from __future__ import annotations

import logging
import socket
import socketserver
from multiprocessing import Queue

from .header import OperationKey
from .header import PacketHeader
from .header import PacketHeaderGenerator
from sigmadsp.sigmastudio.common import CONNECTION_CLOSED
from sigmadsp.sigmastudio.common import ReadRequest
from sigmadsp.sigmastudio.common import ReadResponse
from sigmadsp.sigmastudio.common import SafeloadRequest
from sigmadsp.sigmastudio.common import WriteRequest

# A logger for this module
logger = logging.getLogger(__name__)


class Packet:
    """A packet for data exchange with SigmaStudio."""

    def __init__(self, header: PacketHeader, payload: bytes | None = None):
        """Initialize a new packet from a header and payload.

        Args:
            header (PacketHeader): The packet header object.
            payload (bytes | None): An optional payload. Defaults to None.
        """
        self._header = header

        if payload is not None:
            self.payload = payload

    @property
    def header(self) -> PacketHeader:
        """The header object of the packet."""
        return self._header

    @property
    def payload(self) -> bytes:
        """The packet payload, if any."""
        try:
            return self._payload

        except AttributeError:
            # There is no payload for this packet.
            return b""

    @payload.setter
    def payload(self, new_payload: bytes):
        """Set payload for this packet.

        On packet initialization, this only happens, if the packet is supposed to carry payload.

        Args:
            new_payload (bytes): The payload to store in the packet.
        """
        if self.header["data_length"].value == 0:
            self._payload = new_payload
            self.header["data_length"].value = len(self._payload)

        else:
            assert len(new_payload) == self.header["data_length"].value, (
                f"Payload length {len(new_payload)} does not match "
                f'the expected length {self.header["data_length"].value}.'
            )

            self._payload = new_payload

    def as_bytes(self) -> bytes:
        """Get the full packet as a bytes object."""
        return self.header.as_bytes() + self.payload


class SigmaStudioTcpServer(socketserver.TCPServer):
    """The TCP server that is used for communicating with SigmaStudio."""

    allow_reuse_address = True

    def __init__(
        self,
        server_address: tuple[str, int],
        request_handler_type: type[socketserver.BaseRequestHandler],
        packet_header_generator: PacketHeaderGenerator,
        send_queue: Queue,
        receive_queue: Queue,
        raw_receive_queue: Queue | None = None,
        bind_and_activate=True,
    ):
        """Initialize the ThreadedTCPServer with a Pipe for communicating with the TCP server worker thread.

        Args:
            server_address (Tuple[str, int]): The IP address and port of the server.
            request_handler_type (Type[socketserver.BaseRequestHandler]): The class that handles requests.
            packet_header_generator (PacketHeaderGenerator): A generator for packet headers.
            bind_and_activate (bool, optional): Whether to bind and activate the TCP server. Defaults to True.
            send_queue (Queue): The queue for data to send to SigmaStudio.
            receive_queue (Queue): The queue for data that was received by SigmaStudio.
            raw_receive_queue (Queue | None): The queue for raw binary data, as received by SigmaStudio.
                Used for debugging. Optional, defaults to None.
        """
        self.send_queue = send_queue
        self.receive_queue = receive_queue
        self.raw_receive_queue = raw_receive_queue
        self.packet_header_generator = packet_header_generator

        super().__init__(server_address, request_handler_type, bind_and_activate=bind_and_activate)


class SigmaStudioRequestHandler(socketserver.BaseRequestHandler):
    """Request handler for messages from SigmaStudio."""

    request: socket.socket
    server: SigmaStudioTcpServer

    def read(self, amount: int) -> bytes:
        """Reads the specified amount of data from the socket.

        Args:
            amount (int): The number of bytes to get.

        Returns:
            bytes: The received data.
        """
        data = bytearray()

        while amount > len(data):
            # Wait until the complete TCP payload was received.
            received = self.request.recv(amount - len(data))

            if len(received) == 0:
                # Give up, if no more data arrives.
                # Close the socket.
                self.request.shutdown(socket.SHUT_RDWR)
                self.request.close()

                raise ConnectionError

            data.extend(received)

        return bytes(data)

    def handle_write_request(self, packet: Packet):
        """Handle requests, where SigmaStudio wants to write to the DSP.

        Args:
            packet (Packet): The request packet object.
        """
        request: WriteRequest

        if packet.header.is_safeload:
            logger.debug(
                "[safeload] %s bytes to address 0x%04x",
                packet.header["data_length"].value,
                packet.header["address"].value,
            )
            request = SafeloadRequest(packet.header["address"].value, packet.payload)

        else:
            logger.debug(
                "[write] %s bytes to address 0x%04x", packet.header["data_length"].value, packet.header["address"].value
            )
            request = WriteRequest(packet.header["address"].value, packet.payload)

        self.server.receive_queue.put(request)

    def handle_read_request(self, packet: Packet):
        """Handle requests, where SigmaStudio wants to read from the DSP.

        Args:
            packet (Packet): The request header object.
        """
        logger.debug(
            "[read] %d bytes from address 0x%04x", packet.header["data_length"].value, packet.header["address"].value
        )

        # Notify application of read request.
        self.server.receive_queue.put(ReadRequest(packet.header["address"].value, packet.header["data_length"].value))

        # Wait for payload data that goes into the read response.
        read_response: ReadResponse = self.server.send_queue.get()

        # Build a response header, based on the content of the request header.
        response_header = self.server.packet_header_generator.new_header_from_operation_key(
            OperationKey.READ_RESPONSE_KEY, template=packet.header
        )

        response_packet = Packet(response_header, read_response.data)
        response_packet.header["success"] = 0

        self.request.sendall(response_packet.as_bytes())

    def new_packet(self) -> Packet:
        """Create a new packet from a request."""
        # First, get the operation key. It is always the first received byte.
        operation_byte: bytes = self.read(1)
        header = self.server.packet_header_generator.new_header_from_operation_byte(operation_byte)

        logger.debug("Received operation key 0x%02x.", header["operation"].value)

        # Read the rest of the header.
        remaining_header_bytes = self.read(header.size - 1)
        header.parse(operation_byte + remaining_header_bytes)

        if header.carries_payload:
            return Packet(header, self.read(header["data_length"].value))

        return Packet(header)

    def handle(self):
        """Called, when the TCP server has to handle a request.

        It never stops, except if the connection is reset.
        """
        while True:
            try:
                packet = self.new_packet()

                if self.server.raw_receive_queue is not None:
                    # For debugging purposes, store the full binary package.
                    self.server.raw_receive_queue.put(packet.as_bytes())

                if packet.header.is_write_request:
                    self.handle_write_request(packet)

                elif packet.header.is_read_request:
                    self.handle_read_request(packet)

            except ConnectionError:
                logger.debug("Connection closed in request handler.")
                self.server.receive_queue.put(CONNECTION_CLOSED)
                break



"""This module contains the common classes for SigmaStudio protocol handling."""
from dataclasses import dataclass


@dataclass(frozen=True)
class WriteRequest:
    """SigmaStudio requests to write data to the DSP."""

    address: int
    data: bytes


class SafeloadRequest(WriteRequest):
    """SigmaStudio requests to write data to the DSP using safeload."""


@dataclass(frozen=True)
class ReadRequest:
    """SigmaStudio requests to read data from the DSP."""

    address: int
    length: int


@dataclass(frozen=True)
class ReadResponse:
    """SigmaStudio is sent a response to its ReadRequest."""

    data: bytes


# SigmaStudio disconnected from the server.
CONNECTION_CLOSED = None



"""This module provides functionality for controlling SigmaDSP ADAU1x01 hardware."""
from __future__ import annotations

import logging
import math

from sigmadsp.dsp.common import Dsp
from sigmadsp.helper.conversion import bytes_to_int16
from sigmadsp.helper.conversion import float_to_frac_5_23
from sigmadsp.helper.conversion import frac_5_23_to_float
from sigmadsp.helper.conversion import int16_to_bytes
from sigmadsp.sigmastudio.adau1x01 import Adau1x01HeaderGenerator

# A logger for this module
logger = logging.getLogger(__name__)


class Adau1x0x(Dsp):
    """A class for controlling functionality of Analog Devices Sigma DSPs, especially ADAU1x01 series parts."""

    # Addresses and sizes of important registers
    DSP_CORE_CONTROL_REGISTER = 0x081C
    DSP_CORE_CONTROL_REGISTER_LENGTH = 2

    DSP_CORE_CONTROL_REGISTER_IST_POS = 5
    DSP_CORE_CONTROL_REGISTER_IST_MASK = 1 << DSP_CORE_CONTROL_REGISTER_IST_POS

    # Safeload registers (address, data)
    SAFELOAD_REGISTERS: list[tuple[int, int]] = [
        (0x0815, 0x810),
        (0x0816, 0x811),
        (0x0815, 0x812),
        (0x0818, 0x813),
        (0x0819, 0x814),
    ]

    # Safeload register length
    SAFELOAD_ADDRESS_REGISTER_LENGTH = 2
    SAFELOAD_DATA_REGISTER_LENGTH = 5

    header_generator = Adau1x01HeaderGenerator()

    @staticmethod
    def float_to_frac(value: float) -> int:
        """The method that converts floating-point values to fractional integers on the Adau1x01.

        Args:
            value (float): The floating-point value.

        Returns:
            float: The fractional value.
        """
        return float_to_frac_5_23(value)

    @staticmethod
    def frac_to_float(value: int) -> float:
        """The method that converts fractional integers to floating-point values on the Adau1x01.

        Args:
            value (float): The fractional value.

        Returns:
            float: The float value.
        """
        return frac_5_23_to_float(value)

    def soft_reset(self):
        """Soft reset the DSP.

        Not available on ADAU1x01
        """
        logger.info("Soft-reset is not available on ADAU1x01")

    def safeload(self, address: int, data: bytes):
        """Write data to the chip using hardware safeload.

        Args:
            address (int): Address to write to
            data (bytes): Data to write; multiple words should be concatenated.
        """
        # Number of words with length FIXPOINT_REGISTER_LENGTH in data.
        word_count = math.ceil(len(data) / self.FIXPOINT_REGISTER_LENGTH)

        assert address >= 0

        # load up the address and data in safeload registers
        for sd in range(0, word_count):
            address_register, data_register = Adau1x0x.SAFELOAD_REGISTERS[sd]
            address_bytes = int16_to_bytes(address)
            data_buf = bytearray(Adau1x0x.SAFELOAD_DATA_REGISTER_LENGTH)

            data_buf[1:] = data[sd * Adau1x0x.FIXPOINT_REGISTER_LENGTH : (sd + 1) * Adau1x0x.FIXPOINT_REGISTER_LENGTH]

            self.write(address_register, address_bytes)
            self.write(data_register, data_buf)

        control_bytes = self.read(Adau1x0x.DSP_CORE_CONTROL_REGISTER, Adau1x0x.DSP_CORE_CONTROL_REGISTER_LENGTH)
        control_reg = bytes_to_int16(control_bytes)

        control_reg |= Adau1x0x.DSP_CORE_CONTROL_REGISTER_IST_MASK

        # start safe load
        self.write(Adau1x0x.DSP_CORE_CONTROL_REGISTER, int16_to_bytes(control_reg))



"""This module implements an I2C handler that talks to Sigma DSP devices."""
import logging

from smbus2 import i2c_msg
from smbus2 import SMBus

from .common import DspProtocol
from sigmadsp.helper.conversion import int16_to_bytes

# A logger for this module
logger = logging.getLogger(__name__)


class I2cProtocol(DspProtocol):
    """Handle I2C transfers from and to SigmaDSP chipsets.

    Tested with ADAU1701
    """

    def __init__(self, bus_id: int = 1, device_address: int = 0x38):
        """Initialize the I2C hardware.

        Bus will be 1 for RaspberryPi hardware.

        Args:
            bus_id (int, optional): Bus identifier. Defaults to 1.
            device_address (int, optional): Device address. Defaults to 0x38 (ADAU145x default address).
        """
        self._bus_id = bus_id
        self._device_address = device_address

        self.run()

    def _read(self, address: int, length: int) -> bytes:
        """Read data over the I2C port from a SigmaDSP.

        Args:
            address (int): Address to read from.
            length (int): Number of bytes to read.

        Returns:
            bytes: Data that was read from the DSP.
        """
        address_bytes = list(int16_to_bytes(address))

        write_message = i2c_msg.write(self._device_address, address_bytes)
        read_message = i2c_msg.read(self._device_address, length)

        with SMBus(self._bus_id) as bus:
            bus.i2c_rdwr(write_message, read_message)

        return bytes(read_message)

    def _write(self, address: int, data: bytes) -> int:
        """Write data over the I2C port onto a SigmaDSP.

        Args:
            address (int): Address to write to.
            data (bytes): Data to write.

        Returns:
            int: Number of written bytes.
        """
        payload = list(int16_to_bytes(address))
        payload.extend(list(data))

        write_message = i2c_msg.write(self._device_address, payload)

        with SMBus(self._bus_id) as bus:
            bus.i2c_rdwr(write_message)

        return len(data)



"""Header generator for ADAU1x01 parts."""
from .header import Field
from .header import PacketHeader
from .header import PacketHeaderGenerator


class Adau1x01HeaderGenerator(PacketHeaderGenerator):
    """Header generator for ADAU1x01 parts."""

    @staticmethod
    def _new_write_header() -> PacketHeader:
        """Generate a new header for a ADAU1x01 write packet."""
        return PacketHeader(
            [
                Field("operation", 0, 1),
                Field("safeload", 1, 1),
                Field("channel", 2, 1),
                Field("total_length", 3, 2),
                Field("chip_address", 5, 1),
                Field("data_length", 6, 2),
                Field("address", 8, 2),
            ]
        )

    @staticmethod
    def _new_read_request_header() -> PacketHeader:
        """Generate a new header for a ADAU1x01 read request packet."""
        return PacketHeader(
            [
                Field("operation", 0, 1),
                Field("total_length", 1, 2),
                Field("chip_address", 3, 1),
                Field("data_length", 4, 2),
                Field("address", 6, 2),
            ]
        )

    @staticmethod
    def _new_read_response_header() -> PacketHeader:
        """Generate a new header for a ADAU1x01 read response packet."""
        return PacketHeader(
            [
                Field("operation", 0, 1),
                Field("data_length", 1, 2),
                Field("success", 3, 1),
            ]
        )



"""This module includes many conversion functions that can be used for different purposes.

- Conversion between linear and dB-scale values
- Conversion from bytes-like objects of varying length to integers
- Conversion from integers to bytes-like objects of specified length
"""
from __future__ import annotations

import math
from typing import Literal

BIT_LENGTH_8_24 = 31
BIT_LENGTH_5_23 = 27
SIGMADSP_ENDIANNESS: Literal["big", "little"] = "big"


def clamp(value: float, min_value: float, max_value: float) -> float:
    """Clamps a value to a specified range.

    Args:
        value (float): The value to clamp.
        min_value (float): Lower clamping boundary.
        max_value (float): Upper clamping boundary.

    Returns:
        float: The clamped value.
    """
    if max_value < min_value:
        raise ValueError("Invalid clamping interval [{min_value}, {max_value}].")

    if value > max_value:
        value = max_value

    elif value < min_value:
        value = min_value

    return value


def frac_8_24_to_float(value: int) -> float:
    """Convert a value in the DSPs 32 bit 8.24 fractional representation to float.

    32 bit values consist of 8 integer and 24 fractional bits and are signed.

    Args:
        value (int): Fractional value to convert

    Returns:
        float: Output in float format
    """
    if value.bit_length() > BIT_LENGTH_8_24:
        raise OverflowError

    return value / 2**24


def float_to_frac_8_24(value: float) -> int:
    """Convert a float value to the DSPs 32 bit 8.24 fractional representation.

    32 bit values consist of 8 integer and 24 fractional bits and are signed.

    Args:
        value (float): Float value to convert

    Returns:
        int: Output in DSP fractional format
    """
    frac = int(value * 2**24)
    if frac.bit_length() > BIT_LENGTH_8_24:
        raise OverflowError

    return frac


def frac_5_23_to_float(value: int) -> float:
    """Convert a value in the DSPs 28 bit 5.23 fractional representation to float.

    28 bit values consist of 5 integer and 23 fractional bits and are signed.

    Args:
        value (int): Fractional value to convert

    Returns:
        float: Output in float format
    """
    if value.bit_length() > BIT_LENGTH_5_23:
        raise OverflowError

    return value / 2**23


def float_to_frac_5_23(value: float) -> int:
    """Convert a float value to the DSPs 28 bit 5.23 fractional representation.

    28 bit values consist of 5 integer and 23 fractional bits and are signed.

    Args:
        value (float): Float value to convert

    Returns:
        int: Output in DSP fractional format
    """
    frac = int(value * 2**23)
    if frac.bit_length() > BIT_LENGTH_5_23:
        raise OverflowError

    return frac


def db_to_linear(value_db: float) -> float:
    """Convert a dB-scale value (e.g. voltage) to a linear-scale value.

    Args:
        value_db (float): Input dB value

    Returns:
        float: Output linear value
    """
    return 10 ** (value_db / 20)


def linear_to_db(value_linear: float) -> float:
    """Convert a linear-scale value to a dB-scale value (e.g. voltage).

    Args:
        value_linear (float): The linear input value

    Returns:
        float: Output in dB scale
    """
    if value_linear == 0:
        return -math.inf

    return 20 * math.log10(value_linear)


def bytes_to_int(data: bytes, offset: int = 0, length: int | None = None) -> int:
    """Convert a number of bytes to their integer representation.

    Uses "length" bytes from the "data" input, starting at "offset".

    Args:
        data (bytes): Input bytes
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer
        length (int | None, optional): Number of bytes to convert. Defaults to None,
            where the complete length of data (after offset) is used.

    Returns:
        int: Integer representation of the input data stream.
    """
    if length is not None:
        return int.from_bytes(data[offset : offset + length], byteorder=SIGMADSP_ENDIANNESS)

    return int.from_bytes(data[offset:], byteorder=SIGMADSP_ENDIANNESS)


def bytes_to_int8(data: bytes, offset: int = 0) -> int:
    """Convert one byte to an 8 bit integer value.

    Args:
        data (bytes): Input byte.
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer.

    Returns:
        int: 8 bit integer representation of the input data stream.
    """
    return bytes_to_int(data, offset, length=1)


def bytes_to_int16(data: bytes, offset: int = 0) -> int:
    """Convert two bytes to a 16 bit integer value.

    Args:
        data (bytes): Input bytes.
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer.

    Returns:
        int: 16 bit integer representation of the input data stream.
    """
    return bytes_to_int(data, offset, length=2)


def bytes_to_int32(data: bytes, offset: int = 0) -> int:
    """Convert four bytes to a 32 bit integer value.

    Args:
        data (bytes): Input bytes.
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer.

    Returns:
        int: 32 bit integer representation of the input data stream.
    """
    return bytes_to_int(data, offset, length=4)


def int_to_bytes(value: int, buffer: bytearray | None = None, offset: int = 0, length: int = 1):
    """Fill a buffer with values. If no buffer is provided, a new one is created.

    Args:
        buffer (bytearray | None): The buffer to fill, or None, for creating a new one.
        value (int): The value to pack into the buffer.
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer.
        length (int): Number of bytes to be written.
    """
    if buffer is None:
        buffer = bytearray(length + offset)

    buffer[offset : offset + length] = value.to_bytes(length, byteorder=SIGMADSP_ENDIANNESS)

    return buffer


def int8_to_bytes(value, buffer=None, offset=0):
    """Fill a buffer with an 8 bit value (1 byte). If no buffer is provided, a new one is created.

    Args:
        buffer (bytearray): The buffer to fill
        value (int): The value to pack into the buffer
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer
    """
    return int_to_bytes(value, buffer, offset=offset, length=1)


def int16_to_bytes(value, buffer=None, offset=0):
    """Fill a buffer with a 16 bit value (2 bytes). If no buffer is provided, a new one is created.

    Args:
        value (int): The value to pack into the buffer
        buffer (bytearray, optional): The buffer to fill
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer
    """
    return int_to_bytes(value, buffer, offset=offset, length=2)


def int32_to_bytes(value, buffer=None, offset=0) -> bytearray:
    """Fill a buffer with a 32 bit value (4 bytes). If no buffer is provided, a new one is created.

    Args:
        buffer (bytearray): The buffer to fill
        value (int): The value to pack into the buffer
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer
    """
    return int_to_bytes(value, buffer, offset=offset, length=4)



"""General definitions for interfacing DSPs."""
from __future__ import annotations

import logging
import time
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from dataclasses import field
from math import nan
from typing import Literal

import gpiozero

from sigmadsp.helper.conversion import bytes_to_int32
from sigmadsp.helper.conversion import clamp
from sigmadsp.helper.conversion import db_to_linear
from sigmadsp.helper.conversion import int32_to_bytes
from sigmadsp.helper.conversion import linear_to_db
from sigmadsp.protocols.common import DspProtocol
from sigmadsp.sigmastudio.header import PacketHeaderGenerator

# A logger for this module
logger = logging.getLogger(__name__)

ParameterType = Literal["float", "int"]


class SafetyCheckError(Exception):
    """Custom exception for failed DSP safety checks."""


class ConfigurationError(Exception):
    """Custom exception for an invalid DSP configuration."""


@dataclass
class Pin:
    """A class that describes a general DSP pin."""

    name: str
    number: int

    def __post_init__(self):
        """Initialize the generic device, based on the configured parameters."""
        self.control = gpiozero.GPIODevice(self.number)


@dataclass
class InputPin(Pin):
    """A class that describes a DSP input pin."""

    pull_up: bool
    active_state: bool
    bounce_time: float | None

    def __post_init__(self):
        """Initialize the input device, based on the configured parameters."""
        self.control = gpiozero.DigitalInputDevice(self.number, self.pull_up, self.active_state, self.bounce_time)


@dataclass
class OutputPin(Pin):
    """A class that describes a DSP output pin."""

    initial_value: bool
    active_high: bool

    def __post_init__(self):
        """Initialize the output device, based on the configured parameters."""
        self.control = gpiozero.DigitalOutputDevice(self.number, self.active_high, self.initial_value)


# see https://github.com/python/mypy/issues/5374
@dataclass  # type: ignore
class Dsp(ABC):
    """A generic DSP class, to be extended by child classes."""

    use_safeload: bool
    dsp_protocol: DspProtocol
    pins: list[Pin] = field(default_factory=list)

    # All fixpoint (parameter) registers are four bytes long
    FIXPOINT_REGISTER_LENGTH = 4

    @property
    @abstractmethod
    def header_generator(self) -> PacketHeaderGenerator:
        """The packet header generator of the DSP."""

    @staticmethod
    @abstractmethod
    def float_to_frac(value: float) -> int:
        """The method that converts floating-point values to fractional integers on this Dsp."""

    @staticmethod
    @abstractmethod
    def frac_to_float(value: int) -> float:
        """The method that converts fractional integers to floating-point values on this Dsp."""

    def get_pin_by_name(self, name: str) -> Pin | None:
        """Get a pin by its name.

        Args:
            name (str): The name of the pin.

        Returns:
            Pin | None: The pin, if one matches, None otherwise.
        """
        for pin in self.pins:
            if pin.name == name:
                return pin

        return None

    def has_pin(self, pin: Pin) -> bool:
        """Check, if a pin is known in the list of pins.

        Args:
            pin (Pin): The pin to look for.

        Returns:
            bool: True, if the pin is known, False otherwise.
        """
        return bool(self.get_pin_by_name(pin.name) is not None)

    def add_pin(self, pin: Pin):
        """Add a pin to the list of pins, if it doesn't exist yet.

        Args:
            pin (Pin): The pin to add.
        """
        if not self.has_pin(pin):
            logger.info("Found DSP pin definition '%s' (%d)", pin.name, pin.number)
            self.pins.append(pin)

    def remove_pin_by_name(self, name: str):
        """Remove a pin, based on its name.

        Args:
            name (str): The pin name.
        """
        pin = self.get_pin_by_name(name)

        if pin:
            self.pins.remove(pin)

    def hard_reset(self, delay: float = 0):
        """Hard reset the DSP.

        Set and release the corresponding pin for resetting.
        """
        pin = self.get_pin_by_name("reset")

        if pin:
            logger.info("Hard-resetting the DSP.")

            pin.control.on()
            time.sleep(delay)
            pin.control.off()

        else:
            logger.warning("No hard-reset pin is defined, not resetting.")

        # Soft-reset in the end, for flushing registers.
        self.soft_reset()

    def write(self, address: int, data: bytes):
        """Write data to the DSP using the configured communication handler.

        Args:
            address (int): Address to write to
            data (bytes): Data to write
        """
        assert address >= 0

        self.dsp_protocol.write(address, data)

    def read(self, address: int, length: int) -> bytes:
        """Write data to the DSP using the configured communication handler.

        Args:
            address (int): Address to write to
            length (int): Number of bytes to read
        """
        assert address >= 0
        assert length > 0

        return self.dsp_protocol.read(address, length)

    def set_volume(self, volume_db: float, address: int, relative: bool = False) -> float:
        """Set the volume register at the given address to a certain value in dB.

        Args:
            volume_db (float): The volume setting in dB
            address (int): The volume adjustment register address
            relative (bool): If True, the current DSP register value is changed by ``volume_db``. If False, it is set.

        Returns:
            float: The new volume in dB.
        """
        # Read current volume first
        current_volume_linear = self.get_parameter_value(address, data_format="float")

        if not isinstance(current_volume_linear, float):
            raise TypeError(f"Current volume readout has incorrect type {type(current_volume_linear)}.")

        current_volume_db = linear_to_db(current_volume_linear)

        if volume_db is nan:
            logger.info("Volume value is nan.")
            return current_volume_db

        new_volume_value_db = volume_db + current_volume_db if relative else volume_db

        try:
            # Clamp set volume to safe levels
            new_volume_value_linear = clamp(db_to_linear(new_volume_value_db), 0, 1)

        except OverflowError:
            logger.info("Volume adjustment was too large, ignoring.")
            return current_volume_db

        try:
            self.set_parameter_value(new_volume_value_linear, address, data_format="float")

        except ValueError:
            # Parameter conversion failed.
            return current_volume_db

        logger.info("Set volume to %.2f dB.", linear_to_db(new_volume_value_linear))

        # Read back current volume and verify.
        new_volume_value_linear_from_dsp = self.get_parameter_value(address, data_format="float")

        if not isinstance(new_volume_value_linear_from_dsp, float):
            raise TypeError(f"New volume readout has incorrect type {type(new_volume_value_linear_from_dsp)}.")

        return linear_to_db(new_volume_value_linear_from_dsp)

    def adjust_volume(self, adjustment_db: float, address: int) -> float:
        """Adjust the volume register at the given address by a certain value in dB.

        Args:
            adjustment_db (float): The volume adjustment in dB
            address (int): The volume adjustment register address

        Returns:
            float: The new volume in dB.
        """
        return self.set_volume(adjustment_db, address, relative=True)

    @abstractmethod
    def soft_reset(self):
        """Soft reset the DSP."""

    @abstractmethod
    def safeload(self, address: int, data: bytes):
        """Write data to the chip using chip-specific safeload.

        Args:
            address (int): Address to write to
            data (bytes): Data to write
        """

    def set_parameter_value(self, value: float | int, address: int, data_format: ParameterType) -> None:
        """Set a parameter value for a chosen register address. Registers are 32 bits wide.

        Args:
            value (float | int): The value to store in the register
            address (int): The target address
            data_format (PARAMETER_TYPE): The data type of value. Can be ``float`` or ``int``.
        """
        data_register: bytes | None = None

        if data_format == "float":
            data_register = int32_to_bytes(self.float_to_frac(value))

        elif data_format == "int":
            data_register = int32_to_bytes(value)

        if data_register is not None:
            if self.use_safeload:
                self.safeload(address, data_register)

            else:
                self.write(address, data_register)

    def get_parameter_value(self, address: int, data_format: Literal["int", "float"]) -> float | int | None:
        """Get a parameter value from a chosen register address.

        Args:
            address (int): The address to look at.
            data_format (Literal["int", "float"]): The data type to return the register in. Can be ``float`` or ``int``.

        Returns:
            float | int | None: Representation of the register content in the specified format.
        """
        data_register = self.read(address, Dsp.FIXPOINT_REGISTER_LENGTH)
        data_integer = bytes_to_int32(data_register)

        if data_format == "int":
            return data_integer

        if data_format == "float":
            return self.frac_to_float(data_integer)

        return None



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class ChangeVolume(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_TOKENS_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    RELATIVE_FIELD_NUMBER: builtins.int
    @property
    def name_tokens(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    value: builtins.float
    relative: builtins.bool
    def __init__(
        self,
        *,
        name_tokens: collections.abc.Iterable[builtins.str] | None = ...,
        value: builtins.float = ...,
        relative: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["name_tokens", b"name_tokens", "relative", b"relative", "value", b"value"]) -> None: ...

global___ChangeVolume = ChangeVolume

class ControlParameterRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CHANGE_VOLUME_FIELD_NUMBER: builtins.int
    @property
    def change_volume(self) -> global___ChangeVolume: ...
    def __init__(
        self,
        *,
        change_volume: global___ChangeVolume | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["change_volume", b"change_volume", "command", b"command"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["change_volume", b"change_volume", "command", b"command"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["command", b"command"]) -> typing_extensions.Literal["change_volume"] | None: ...

global___ControlParameterRequest = ControlParameterRequest

class LoadParameters(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONTENT_FIELD_NUMBER: builtins.int
    @property
    def content(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        content: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["content", b"content"]) -> None: ...

global___LoadParameters = LoadParameters

class Register(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ADDRESS_FIELD_NUMBER: builtins.int
    LENGTH_FIELD_NUMBER: builtins.int
    DATA_FIELD_NUMBER: builtins.int
    address: builtins.int
    length: builtins.int
    @property
    def data(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    def __init__(
        self,
        *,
        address: builtins.int = ...,
        length: builtins.int = ...,
        data: collections.abc.Iterable[builtins.bytes] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["address", b"address", "data", b"data", "length", b"length"]) -> None: ...

global___Register = Register

class ControlRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RESET_DSP_FIELD_NUMBER: builtins.int
    HARD_RESET_DSP_FIELD_NUMBER: builtins.int
    LOAD_PARAMETERS_FIELD_NUMBER: builtins.int
    READ_REGISTER_FIELD_NUMBER: builtins.int
    WRITE_REGISTER_FIELD_NUMBER: builtins.int
    reset_dsp: builtins.bool
    hard_reset_dsp: builtins.bool
    @property
    def load_parameters(self) -> global___LoadParameters: ...
    @property
    def read_register(self) -> global___Register: ...
    @property
    def write_register(self) -> global___Register: ...
    def __init__(
        self,
        *,
        reset_dsp: builtins.bool = ...,
        hard_reset_dsp: builtins.bool = ...,
        load_parameters: global___LoadParameters | None = ...,
        read_register: global___Register | None = ...,
        write_register: global___Register | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["command", b"command", "hard_reset_dsp", b"hard_reset_dsp", "load_parameters", b"load_parameters", "read_register", b"read_register", "reset_dsp", b"reset_dsp", "write_register", b"write_register"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["command", b"command", "hard_reset_dsp", b"hard_reset_dsp", "load_parameters", b"load_parameters", "read_register", b"read_register", "reset_dsp", b"reset_dsp", "write_register", b"write_register"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["command", b"command"]) -> typing_extensions.Literal["reset_dsp", "hard_reset_dsp", "load_parameters", "read_register", "write_register"] | None: ...

global___ControlRequest = ControlRequest

class ControlResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SUCCESS_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    success: builtins.bool
    message: builtins.str
    def __init__(
        self,
        *,
        success: builtins.bool = ...,
        message: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["message", b"message", "success", b"success"]) -> None: ...

global___ControlResponse = ControlResponse



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: control.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rcontrol.proto\x12\x18sigmadsp.backend_service\"D\n\x0c\x43hangeVolume\x12\x13\n\x0bname_tokens\x18\x01 \x03(\t\x12\r\n\x05value\x18\x02 \x01(\x01\x12\x10\n\x08relative\x18\x03 \x01(\x08\"e\n\x17\x43ontrolParameterRequest\x12?\n\rchange_volume\x18\x01 \x01(\x0b\x32&.sigmadsp.backend_service.ChangeVolumeH\x00\x42\t\n\x07\x63ommand\"!\n\x0eLoadParameters\x12\x0f\n\x07\x63ontent\x18\x01 \x03(\t\"9\n\x08Register\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\r\x12\x0e\n\x06length\x18\x02 \x01(\r\x12\x0c\n\x04\x64\x61ta\x18\x03 \x03(\x0c\"\x8a\x02\n\x0e\x43ontrolRequest\x12\x13\n\treset_dsp\x18\x01 \x01(\x08H\x00\x12\x18\n\x0ehard_reset_dsp\x18\x02 \x01(\x08H\x00\x12\x43\n\x0fload_parameters\x18\x03 \x01(\x0b\x32(.sigmadsp.backend_service.LoadParametersH\x00\x12;\n\rread_register\x18\x04 \x01(\x0b\x32\".sigmadsp.backend_service.RegisterH\x00\x12<\n\x0ewrite_register\x18\x05 \x01(\x0b\x32\".sigmadsp.backend_service.RegisterH\x00\x42\t\n\x07\x63ommand\"3\n\x0f\x43ontrolResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t2\xdc\x01\n\x07\x42\x61\x63kend\x12^\n\x07\x63ontrol\x12(.sigmadsp.backend_service.ControlRequest\x1a).sigmadsp.backend_service.ControlResponse\x12q\n\x11\x63ontrol_parameter\x12\x31.sigmadsp.backend_service.ControlParameterRequest\x1a).sigmadsp.backend_service.ControlResponseb\x06proto3')



_CHANGEVOLUME = DESCRIPTOR.message_types_by_name['ChangeVolume']
_CONTROLPARAMETERREQUEST = DESCRIPTOR.message_types_by_name['ControlParameterRequest']
_LOADPARAMETERS = DESCRIPTOR.message_types_by_name['LoadParameters']
_REGISTER = DESCRIPTOR.message_types_by_name['Register']
_CONTROLREQUEST = DESCRIPTOR.message_types_by_name['ControlRequest']
_CONTROLRESPONSE = DESCRIPTOR.message_types_by_name['ControlResponse']
ChangeVolume = _reflection.GeneratedProtocolMessageType('ChangeVolume', (_message.Message,), {
  'DESCRIPTOR' : _CHANGEVOLUME,
  '__module__' : 'control_pb2'
  # @@protoc_insertion_point(class_scope:sigmadsp.backend_service.ChangeVolume)
  })
_sym_db.RegisterMessage(ChangeVolume)

ControlParameterRequest = _reflection.GeneratedProtocolMessageType('ControlParameterRequest', (_message.Message,), {
  'DESCRIPTOR' : _CONTROLPARAMETERREQUEST,
  '__module__' : 'control_pb2'
  # @@protoc_insertion_point(class_scope:sigmadsp.backend_service.ControlParameterRequest)
  })
_sym_db.RegisterMessage(ControlParameterRequest)

LoadParameters = _reflection.GeneratedProtocolMessageType('LoadParameters', (_message.Message,), {
  'DESCRIPTOR' : _LOADPARAMETERS,
  '__module__' : 'control_pb2'
  # @@protoc_insertion_point(class_scope:sigmadsp.backend_service.LoadParameters)
  })
_sym_db.RegisterMessage(LoadParameters)

Register = _reflection.GeneratedProtocolMessageType('Register', (_message.Message,), {
  'DESCRIPTOR' : _REGISTER,
  '__module__' : 'control_pb2'
  # @@protoc_insertion_point(class_scope:sigmadsp.backend_service.Register)
  })
_sym_db.RegisterMessage(Register)

ControlRequest = _reflection.GeneratedProtocolMessageType('ControlRequest', (_message.Message,), {
  'DESCRIPTOR' : _CONTROLREQUEST,
  '__module__' : 'control_pb2'
  # @@protoc_insertion_point(class_scope:sigmadsp.backend_service.ControlRequest)
  })
_sym_db.RegisterMessage(ControlRequest)

ControlResponse = _reflection.GeneratedProtocolMessageType('ControlResponse', (_message.Message,), {
  'DESCRIPTOR' : _CONTROLRESPONSE,
  '__module__' : 'control_pb2'
  # @@protoc_insertion_point(class_scope:sigmadsp.backend_service.ControlResponse)
  })
_sym_db.RegisterMessage(ControlResponse)

_BACKEND = DESCRIPTOR.services_by_name['Backend']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CHANGEVOLUME._serialized_start=43
  _CHANGEVOLUME._serialized_end=111
  _CONTROLPARAMETERREQUEST._serialized_start=113
  _CONTROLPARAMETERREQUEST._serialized_end=214
  _LOADPARAMETERS._serialized_start=216
  _LOADPARAMETERS._serialized_end=249
  _REGISTER._serialized_start=251
  _REGISTER._serialized_end=308
  _CONTROLREQUEST._serialized_start=311
  _CONTROLREQUEST._serialized_end=577
  _CONTROLRESPONSE._serialized_start=579
  _CONTROLRESPONSE._serialized_end=630
  _BACKEND._serialized_start=633
  _BACKEND._serialized_end=853
# @@protoc_insertion_point(module_scope)



"""Factory module for generating DSP objects."""
from __future__ import annotations

import logging

from .adau14xx import Adau14xx
from .adau1x0x import Adau1x0x
from .common import ConfigurationError
from .common import Dsp
from .common import InputPin
from .common import OutputPin
from sigmadsp.protocols.i2c import I2cProtocol
from sigmadsp.protocols.spi import SpiProtocol

logger = logging.getLogger(__name__)

ADAU_14XX = "adau14xx"
ADAU_1X0X = "adau1x0x"


def dsp_factory(dsp_type_name: str) -> type[Dsp]:
    """Creates Dsp objects, based on the name of the DSP.

    Args:
        dsp_type_name (str): The name of the DSP.

    Returns:
        type[Dsp]: The matching DSP class.
    """
    if dsp_type_name == ADAU_14XX:
        return Adau14xx

    if dsp_type_name == ADAU_1X0X:
        return Adau1x0x

    raise TypeError("DSP type {dsp_type} is not known.")


def dsp_from_config(config: dict) -> Dsp:
    """Parse a configuration dictionary and create a new DSP from it.

    Args:
        config (dict): The configuration dictionary.

    Returns:
        Dsp: The created DSP object.
    """
    dsp: Dsp

    try:
        dsp_protocol_name: str = config["dsp"]["protocol"].lower()
        dsp_type_name: str = config["dsp"]["type"].lower()
        bus = int(config["dsp"]["bus_number"])
        device = int(config["dsp"]["device_address"])

    except KeyError as e:
        logger.error("Key %s missing from the DSP configuration.", e.args[0])
        raise ConfigurationError from e

    else:
        # Generate the protocol first (e.g. SPI or I2C).
        dsp_protocol: SpiProtocol | I2cProtocol

        if dsp_protocol_name == "spi":
            dsp_protocol = SpiProtocol(bus=bus, device=device)

        elif dsp_protocol_name == "i2c":
            dsp_protocol = I2cProtocol(bus_id=bus, device_address=device)

        else:
            raise TypeError(f"Unknown DSP protocol {dsp_protocol_name}.")

        # Then, generate the dsp itself.
        dsp_type: type[Dsp] = dsp_factory(dsp_type_name)

        # Use safeload by default.
        dsp = dsp_type(True, dsp_protocol)

    try:
        # Parse the configuration for pin definitions.
        for pin_definition_key in config["dsp"]["pins"]:
            pin_definition = config["dsp"]["pins"][pin_definition_key]

            if pin_definition["mode"] == "output":
                output_pin = OutputPin(
                    pin_definition_key,
                    pin_definition["number"],
                    pin_definition["initial_state"],
                    pin_definition["active_high"],
                )

                dsp.add_pin(output_pin)

            elif pin_definition["mode"] == "input":
                input_pin = InputPin(
                    pin_definition_key,
                    pin_definition["number"],
                    pin_definition["pull_up"],
                    pin_definition["active_state"],
                    pin_definition["bounce_time"],
                )

                dsp.add_pin(input_pin)

    except (KeyError, TypeError):
        logger.info("No DSP pin definitions were found in the configuration file.")

    return dsp



"""This module is the frontend to the SigmaDSP backend service.

It can control the backend via the command line.
"""
from __future__ import annotations

import logging
from io import TextIOWrapper
from ipaddress import IPv4Address

import click
import grpc

from sigmadsp.generated.backend_service.control_pb2 import ControlParameterRequest
from sigmadsp.generated.backend_service.control_pb2 import ControlRequest
from sigmadsp.generated.backend_service.control_pb2 import ControlResponse
from sigmadsp.generated.backend_service.control_pb2_grpc import BackendStub


class Channel:
    """Opens a channel to and handles control of the backend."""

    stub = BackendStub

    def __init__(self, backend_address: IPv4Address, backend_port: int):
        """Initialize the channel.

        Args:
            backend_address (ip_address): The IP address of the backend.
            backend_port (int): The backend port.
        """
        self.address = f"{backend_address}:{backend_port}"

    def _emit(self, request: ControlRequest | ControlParameterRequest):
        """Emits a request towards the backend.

        During emission, the channel is opened.

        Args:
            request (ControlRequest | ControlParameterRequest): The request to emit.
        """
        with grpc.insecure_channel(self.address) as channel:
            stub = BackendStub(channel)
            response: ControlResponse | None = None

            if isinstance(request, ControlRequest):
                response = stub.control(request)

            elif isinstance(request, ControlParameterRequest):
                response = stub.control_parameter(request)

            if response is not None:
                logging.info(response and response.message)

    def read_register(self, address: int, length: int):
        """Read a DSP register.

        Args:
            address (int): The address to read from.
            length (int): The length of data to read in bytes.
        """
        request = ControlRequest()
        request.read_register.address = address
        request.read_register.length = length
        self._emit(request)

    def change_volume(self, volume: float, relative: bool):
        """Change the DSP volume by a certain amount.

        Args:
            volume (float): The volume change in dB.
            relative (bool): If True, performs a relative change. If False, sets the provided volume.
        """
        request = ControlParameterRequest()
        request.change_volume.name_tokens[:] = ["main"]
        request.change_volume.value = volume
        request.change_volume.relative = relative
        self._emit(request)

    def reset(self, hard: bool):
        """Reset the DSP.

        Args:
            hard (bool): If True, performs a hard reset. If False, performs a soft reset.
        """
        request = ControlRequest()

        if hard:
            request.hard_reset_dsp = True

        else:
            request.reset_dsp = True

        self._emit(request)

    def load_parameters(self, file: TextIOWrapper):
        """Load a parameter file to the backend.

        Args:
            file (TextIOWrapper): The file to load.
        """
        request = ControlRequest()
        request.load_parameters.content[:] = file.readlines()
        self._emit(request)


@click.group()
@click.pass_context
@click.option(
    "--port",
    default=50051,
    show_default=True,
    type=int,
    help="Set the port, on which the backend listens for requests.",
)
@click.option(
    "--ip",
    default=IPv4Address("127.0.0.1"),
    show_default=True,
    type=IPv4Address,
    help="Set the IP address, on which the backend listens for requests.",
)
def sigmadsp(ctx: click.Context, port: int, ip: IPv4Address):
    """Command-line tool for controlling the sigmadsp-backend."""
    logging.basicConfig(level=logging.DEBUG)
    ctx.obj = Channel(ip, port)


@sigmadsp.command(context_settings={"ignore_unknown_options": True})
@click.pass_obj
@click.argument("file", type=click.File("r"))
def load_parameters(channel: Channel, file: TextIOWrapper):
    """Load a parameter file."""
    channel.load_parameters(file)


@sigmadsp.command(context_settings={"ignore_unknown_options": True})
@click.pass_obj
@click.argument("volume", type=float)
def set_volume(channel: Channel, volume: float):
    """Sets the volume to a certain value in dB."""
    channel.change_volume(volume, False)


@sigmadsp.command(context_settings={"ignore_unknown_options": True})
@click.pass_obj
@click.argument("change", type=float)
def change_volume(channel: Channel, change: float):
    """Changes the volume by a certain amount in dB."""
    channel.change_volume(change, True)


@sigmadsp.command()
@click.pass_obj
@click.option("--hard", is_flag=True)
def reset(channel: Channel, hard: bool):
    """Resets the DSP."""
    channel.reset(hard)


@sigmadsp.command()
@click.pass_obj
@click.argument("address", type=int)
@click.argument("length", type=int)
def read_register(channel: Channel, address: int, length: int):
    """Reads a DSP register."""
    channel.read_register(address, length)


if __name__ == "__main__":
    sigmadsp()  # pylint: disable=no-value-for-parameter



"""This module provides functionality for controlling SigmaDSP ADAU14xx hardware."""
import logging
import math

from sigmadsp.dsp.common import Dsp
from sigmadsp.helper.conversion import float_to_frac_8_24
from sigmadsp.helper.conversion import frac_8_24_to_float
from sigmadsp.helper.conversion import int16_to_bytes
from sigmadsp.helper.conversion import int32_to_bytes
from sigmadsp.sigmastudio.adau14xx import Adau14xxHeaderGenerator

# A logger for this module
logger = logging.getLogger(__name__)


class Adau14xx(Dsp):
    """A class for controlling functionality of Analog Devices Sigma DSPs, especially ADAU14xx series parts."""

    # Addresses and sizes of important registers
    RESET_REGISTER = 0xF890
    RESET_REGISTER_LENGTH = 2

    # safeload registers - SigmaStudio 3.14 and newer
    SAFELOAD_DATA_REGISTERS = [0x6000, 0x6001, 0x6002, 0x6003, 0x6004]
    SAFELOAD_ADDRESS_REGISTER = 0x6005
    SAFELOAD_COUNT_REGISTER = 0x6006
    SAFELOAD_DATA_REGISTER_LENGTH = 4

    header_generator = Adau14xxHeaderGenerator()

    @staticmethod
    def float_to_frac(value: float) -> int:
        """The method that converts floating-point values to fractional integers on the Adau14xx."""
        return float_to_frac_8_24(value)

    @staticmethod
    def frac_to_float(value: int) -> float:
        """The method that converts fractional integers to floating-point values on the Adau14xx."""
        return frac_8_24_to_float(value)

    def soft_reset(self):
        """Soft reset the DSP.

        Set and release the corresponding register for resetting.
        """
        self.write(Adau14xx.RESET_REGISTER, int16_to_bytes(0))
        self.write(Adau14xx.RESET_REGISTER, int16_to_bytes(1))
        logger.info("Soft-resetting the DSP.")

    def safeload(self, address: int, data: bytes):
        """Write data to the chip using software safeload.

        Args:
            address (int): Address to write to
            data (bytes): Data to write; multiple parameters should be concatenated
        """
        # Number of words with length FIXPOINT_REGISTER_LENGTH in data.
        word_count = math.ceil(len(data) / self.FIXPOINT_REGISTER_LENGTH)

        assert address >= 0

        if word_count > len(Adau14xx.SAFELOAD_DATA_REGISTERS):
            raise IndexError(
                f"Cannot write {word_count * self.FIXPOINT_REGISTER_LENGTH} bytes by means of software safeload, "
                f"the maximum is {len(Adau14xx.SAFELOAD_DATA_REGISTERS) * self.FIXPOINT_REGISTER_LENGTH} bytes."
            )
        for register_index, register_address in zip(range(word_count), Adau14xx.SAFELOAD_DATA_REGISTERS):
            data_buf = data[
                register_index * self.FIXPOINT_REGISTER_LENGTH : (register_index + 1) * self.FIXPOINT_REGISTER_LENGTH
            ]
            self.write(register_address, data_buf)

        # TODO: test if the address is supposed to be shifted down by 1 as old forum posts suggest
        self.write(self.SAFELOAD_ADDRESS_REGISTER, int32_to_bytes(address))
        self.write(self.SAFELOAD_COUNT_REGISTER, int32_to_bytes(word_count))



"""This module implements the base class for I2C and SPI communication handlers."""
import logging
import threading
from abc import ABC
from abc import abstractmethod
from multiprocessing import Queue

from sigmadsp.sigmastudio.common import ReadRequest
from sigmadsp.sigmastudio.common import WriteRequest

# A logger for this module
logger = logging.getLogger(__name__)


class DspProtocol(ABC):
    """Base class for communication handlers talking to SigmaDSP chipsets."""

    transmit_queue: Queue
    receive_queue: Queue

    def run(self):
        """Start the DSP protocol thread."""
        # Generate queues, for communicating with the protocol handler thread within this class.
        self.transmit_queue = Queue()
        self.receive_queue = Queue()

        protocol = self.__class__.__name__

        logger.info("Starting %s handling thread.", protocol)
        self.thread = threading.Thread(target=self.serve_forever, name=f"{protocol} handler thread", daemon=True)
        self.thread.start()

    def write(self, address: int, data: bytes):
        """Write data over the hardware interface via the pipe to the hardware thread.

        Args:
            address (int): DSP register address to write to
            data (bytes): Binary data to write
        """
        self.transmit_queue.put(WriteRequest(address, data))

    def read(self, address: int, length: int) -> bytes:
        """Read data from the hardware interface via the pipe to the hardware thread.

        Args:
            address (int): DSP register address to read from
            length (int): Number of bytes to read

        Returns:
            bytes: Register content
        """
        self.transmit_queue.put(ReadRequest(address, length))
        return self.receive_queue.get()

    def serve_forever(self):
        """Handle incoming requests for writing or reading data."""
        while True:
            try:
                request = self.transmit_queue.get()

            except EOFError:
                break

            if isinstance(request, WriteRequest):
                self._write(request.address, request.data)

            elif isinstance(request, ReadRequest):
                data = self._read(request.address, request.length)

                try:
                    self.receive_queue.put(data)

                except EOFError:
                    break

    @abstractmethod
    def _read(self, address: int, length: int) -> bytes:
        """Read data from a SigmaDSP.

        Args:
            address (int): Address to read from.
            length (int): Number of bytes to read.

        Returns:
            bytes: Data that was read from the DSP.
        """

    @abstractmethod
    def _write(self, address: int, data: bytes) -> int:
        """Write data onto a SigmaDSP.

        Args:
            address (int): Address to write to.
            data (bytes): Data to write.

        Returns:
            int: Number of written bytes.
        """



"""The backend application of sigmadsp.

This module includes the main application that provides an interface between the TCP server that faces SigmaStudio,
and an SPI handler that controls a DSP.

Commands from Sigma Studio are received, and translated to SPI read/write requests.
"""
import argparse
import contextlib
import logging
import os
import sched
import sys
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import Queue
from pathlib import Path
from typing import Callable

import grpc
from retry.api import retry_call

import sigmadsp
from sigmadsp.dsp.common import ConfigurationError
from sigmadsp.dsp.common import Dsp
from sigmadsp.dsp.common import SafetyCheckError
from sigmadsp.dsp.factory import dsp_from_config
from sigmadsp.generated.backend_service.control_pb2 import ControlParameterRequest
from sigmadsp.generated.backend_service.control_pb2 import ControlRequest
from sigmadsp.generated.backend_service.control_pb2 import ControlResponse
from sigmadsp.generated.backend_service.control_pb2_grpc import (
    add_BackendServicer_to_server,
)
from sigmadsp.generated.backend_service.control_pb2_grpc import BackendServicer
from sigmadsp.helper.settings import SigmadspSettings
from sigmadsp.sigmastudio.common import CONNECTION_CLOSED
from sigmadsp.sigmastudio.common import ReadRequest
from sigmadsp.sigmastudio.common import ReadResponse
from sigmadsp.sigmastudio.common import SafeloadRequest
from sigmadsp.sigmastudio.common import WriteRequest
from sigmadsp.sigmastudio.server import SigmaStudioRequestHandler
from sigmadsp.sigmastudio.server import SigmaStudioTcpServer

# A logger for this module
logger = logging.getLogger(__name__)


class BackendService(BackendServicer):
    """The backend service that handles the underlying TCP server and SPI handler.

    This service also reacts to rpyc remote procedure calls, for performing actions with the DSP over SPI.
    """

    dsp: Dsp
    send_queue: Queue
    receive_queue: Queue

    def __init__(
        self,
        settings: SigmadspSettings,
        dsp_from_config_fn: Callable[[dict], Dsp] = dsp_from_config,
    ):
        """Initialize service and start all relevant threads (TCP, SPI).

        Args:
            settings (SigmadspSettings): The settings object.
            dsp_from_config_fn (Callable[[Dict], Dsp]): The function to use for generating new Dsp objects
                from a config. The callable takes the config dictionary as an argument, and creates a new Dsp. Optional,
                defaults to ``dsp_from_config``.
        """
        super().__init__()

        self._active = True
        self.send_queue = Queue()
        self.receive_queue = Queue()

        # If configuration_unlocked is False, no DSP parameters can be changed.
        self.configuration_unlocked: bool = False

        self.settings = settings
        self.config = self.settings.config

        # Create a scheduler for recurring tasks
        self.scheduler = sched.scheduler(time.time, time.sleep)

        try:
            self.dsp = dsp_from_config_fn(self.config)

        except ConfigurationError:
            logger.error("DSP configuration is broken! Aborting")
            sys.exit(1)

        logger.info(
            "Specified DSP type is '%s', using the '%s' protocol.",
            type(self.dsp).__name__,
            type(self.dsp.dsp_protocol).__name__,
        )

        # Create the startup thread
        startup_worker_thread = threading.Thread(target=self.startup_worker, name="Startup worker", daemon=True)
        startup_worker_thread.start()

    def startup_worker(self):
        """A thread that performs a startup check and then starts the rest of the threads."""
        # Reset the DSP first.
        self.dsp.hard_reset()

        try:
            logger.info("Run startup safety check.")
            self.retry_safety_check()

        except SafetyCheckError:
            logger.warning("Startup safety check failed.")

        logger.info("Startup finished.")

        # Create a SigmaTCPServer and run it.
        self.sigma_tcp_server = SigmaStudioTcpServer(
            (self.config["host"]["ip"], self.config["host"]["port"]),
            SigmaStudioRequestHandler,
            self.dsp.header_generator,
            self.send_queue,
            self.receive_queue,
        )

        self.sigma_tcp_server_worker_thread = threading.Thread(
            target=self.sigma_tcp_server_worker,
            daemon=True,
            name="Sigma TCP server worker",
        )
        self.sigma_tcp_server_worker_thread.start()

        logger.info(
            "Sigma TCP server started on [%s]:%d.",
            self.config["host"]["ip"],
            self.config["host"]["port"],
        )

        # Create the request worker thread.
        self.sigma_studio_worker_thread = threading.Thread(
            target=self.sigma_studio_worker, name="SigmaStudio worker", daemon=True
        )
        self.sigma_studio_worker_thread.start()

    def trigger_retry_safety_check(self):
        """Runs the safety check in a new thread."""
        safety_check_thread = threading.Thread(target=self.retry_safety_check, daemon=True)
        safety_check_thread.start()

    def retry_safety_check(self) -> None:
        """The ``safety_check``, but with retries."""
        retry_call(self.safety_check, exceptions=SafetyCheckError, tries=5, delay=5)

    def safety_check(self) -> None:
        """Check a hash cell within the DSP's memory against a hash value from the parameter file.

        Only if they match, configuration of DSP parameters is allowed. Other operations (e.g. reset, or
        loading a new parameter file) are still allowed, even in a locked configuration state.
        """
        self.configuration_unlocked = False

        if not self.settings.parameter_parser:
            logger.warning("No parameter file was loaded! Configuration remains locked.")
            return

        safety_hash_cell = self.settings.parameter_parser.safety_hash_cell

        if safety_hash_cell is None:
            logger.warning("No safety hash cell exists in the DSP configuration! Configuration remains locked.")
            self.configuration_unlocked = False

        else:
            dsp_hash = self.dsp.get_parameter_value(safety_hash_cell.parameter_address, data_format="int")
            logger.info("Safety hash address: 0x%04x.", safety_hash_cell.parameter_address)

            if safety_hash_cell.parameter_value != dsp_hash:
                logger.warning(
                    "Safety hash cell content does not match! Expected %d, but read %d. Configuration remains locked.",
                    safety_hash_cell.parameter_value,
                    dsp_hash,
                )
                self.configuration_unlocked = False
                raise SafetyCheckError

            logger.info("Safety check successful. Configuration unlocked.")
            self.configuration_unlocked = True

    def close(self):
        """Close the backend service and terminate its threads."""
        self._active = False

        self.send_queue.close()
        self.receive_queue.close()

        self.sigma_studio_worker_thread.join()

        self.sigma_tcp_server.shutdown()
        self.sigma_tcp_server_worker_thread.join()

    def sigma_tcp_server_worker(self):
        """Sigma TCP server worker.

        Simply runs the TCP server.
        """
        with self.sigma_tcp_server:
            self.sigma_tcp_server.serve_forever()

        logger.info("Sigma TCP server worker terminated.")

    def sigma_studio_worker(self):
        """Sigma studio worker.

        Gets requests from the TCP server component and forwards them to the SPI handler.
        """
        with self.sigma_tcp_server:
            while self._active:
                try:
                    request = self.receive_queue.get()

                    if request is CONNECTION_CLOSED:
                        ...

                    elif isinstance(request, WriteRequest):
                        self.dsp.write(request.address, request.data)

                    elif isinstance(request, SafeloadRequest):
                        self.dsp.safeload(request.address, request.data)

                    elif isinstance(request, ReadRequest):
                        payload = self.dsp.read(request.address, request.length)
                        self.send_queue.put(ReadResponse(payload))

                    else:
                        raise TypeError(f"Unknown command type {type(request)}.")

                except EOFError:
                    # Access to one of the queues failed, terminate the worker.
                    break

        logger.info("Sigma studio worker terminated.")

    def control_parameter(self, request: ControlParameterRequest, context):
        """Main backend entry point for control messages that change or read parameters.

        Here, the configuration has to be in an unlocked state, which requires a hash number in the DSP firmware to
        match the hash number that is provided in the parameter file.

        Args:
            request (ControlParameterRequest): The request that the backend shall handle.
            context (Any): The context within which to handle the request (unused).

        Returns:
            ControlResponse: A control response message, returned to the caller.
        """
        response = ControlResponse()

        # Determine the type of control request.
        command = request.WhichOneof("command")
        response.success = False

        if not self.settings.parameter_parser:
            response.message = "No parameters were loaded, parameters cannot be changed."
            return response

        if not self.configuration_unlocked:
            # Do not allow to change parameters.
            response.message = "Configuration is currently locked, parameters cannot be changed."
            return response

        if command == "change_volume":
            volume_cells_to_adjust = self.settings.parameter_parser.get_matching_cells_by_name_tokens(
                self.settings.parameter_parser.volume_cells,
                list(request.change_volume.name_tokens),
            )

            if not volume_cells_to_adjust:
                response.message = (
                    f"No volume cell identified by {request.change_volume.name_tokens} was found."
                    "Available controls are "
                    f"{([cell.name_tokens for cell in self.settings.parameter_parser.volume_cells])}."
                )

            for volume_cell in volume_cells_to_adjust:
                if request.change_volume.relative:
                    new_volume_db = self.dsp.adjust_volume(
                        request.change_volume.value,
                        volume_cell.parameter_address,
                    )

                elif not request.change_volume.relative:
                    new_volume_db = self.dsp.set_volume(
                        request.change_volume.value,
                        volume_cell.parameter_address,
                    )

                response.message = f"Set volume of cell '{volume_cell.full_name}' to {new_volume_db:.2f} dB."

        response.success = True
        return response

    def control(self, request: ControlRequest, context):
        """Main backend entry point for control messages.

        Args:
            request (ControlRequest): The request that the backend shall handle.
            context (Any): The context within which to handle the request (unused).

        Returns:
            ControlResponse: A control response message, returned to the caller.
        """
        response = ControlResponse()

        # Determine the type of control request.
        command = request.WhichOneof("command")

        if command == "reset_dsp":
            self.dsp.soft_reset()
            self.trigger_retry_safety_check()

            response.message = "Soft-reset DSP."
            response.success = True

        elif command == "hard_reset_dsp":
            self.dsp.hard_reset()
            self.trigger_retry_safety_check()

            response.message = "Hard-reset DSP."
            response.success = True

        elif command == "load_parameters":
            self.settings.store_parameters(list(request.load_parameters.content))

            # Repeat safety check after loading a new set of parameters
            with contextlib.suppress(SafetyCheckError):
                self.safety_check()

            if self.configuration_unlocked:
                response.message = "Loaded parameters, control is unlocked."
                response.success = True

            else:
                response.message = "Safety check failed, parameters cannot be adjusted."
                response.success = False

        elif command == "read_register":
            value: bytes = self.dsp.read(request.read_register.address, request.read_register.length)

            response.message = f"0x{value.hex()}"
            response.success = True

        else:
            response.message = f"Unknown command '{command}'."
            response.success = False

        return response


def launch(settings: SigmadspSettings):
    """Launch the backend application.

    Args:
        settings (SigmadspSettings): Settings object for the backend application.
    """
    # Create the backend service, a grpc service
    grpc_server = grpc.server(ThreadPoolExecutor(max_workers=10))

    backend_port = settings.config["backend"]["port"]
    backend_address = f"[::]:{backend_port}"

    add_BackendServicer_to_server(BackendService(settings), grpc_server)
    grpc_server.add_insecure_port(backend_address)
    grpc_server.start()

    logger.info("Backend service started on %s", backend_address)

    grpc_server.wait_for_termination()


def main():
    """Launch the backend with default settings."""
    if os.environ.get("SIGMADSP_DEBUG") is None:
        logging.basicConfig(level=logging.INFO)

    else:
        logging.basicConfig(level=logging.DEBUG)

    logger.info("Starting the sigmadsp backend, version %s.", sigmadsp.__version__)

    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument(
        "-s",
        "--settings",
        required=False,
        help=("specifies the settings file path for configuring the sigmadsp backend application"),
    )
    arguments = argument_parser.parse_args()

    settings = SigmadspSettings(Path(arguments.settings))
    launch(settings)


if __name__ == "__main__":
    main()



"""This module describes the header of a packet that is exchanged with SigmaStudio.

Headers contain individual fields that follow each other in a certain sequence.
"""
from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum
from typing import Literal

from sigmadsp.helper.conversion import bytes_to_int
from sigmadsp.helper.conversion import int8_to_bytes
from sigmadsp.helper.conversion import int_to_bytes


class OperationKey(Enum):
    """Possible operation keys that are exchanged with SigmaStudio."""

    READ_REQUEST_KEY = 0x0A
    READ_RESPONSE_KEY = 0x0B
    WRITE_KEY = 0x09


FieldName = Literal[
    "operation", "safeload", "channel", "total_length", "chip_address", "data_length", "address", "success", "reserved"
]


@dataclass
class Field:
    """A class that represents a single field in the header."""

    # The name of the field.
    name: FieldName

    # The offset of the field in bytes from the start of the header.
    offset: int

    # The size of the field in bytes.
    size: int

    @property
    def value(self) -> int:
        """The stored value."""
        return self._value

    @value.setter
    def value(self, new_value: int | bytes | bytearray):
        """Store a new value and convert it before storage, if required.

        Args:
            new_value (int | bytes | bytearray): The new value to set. If ``int``, it is stored without conversion.
                If ``bytes`` or ``bytearray``, it is first converted to int.

        Raises:
            TypeError: If the value type is not supported.
        """
        if isinstance(new_value, bytearray):
            new_value = bytes(new_value)

        if isinstance(new_value, bytes):
            assert len(new_value) == self.size, (
                f"New value {new_value!r} for '{self.name}' does not match expected size: "
                f"expected {self.size}, got {len(new_value)}"
            )
            self._value = bytes_to_int(new_value, 0)

        elif isinstance(new_value, int):
            assert new_value.bit_length() <= self.size * 8, (
                f"New value {new_value} for '{self.name}' exceeds expected size: "
                f"expected <={self.size * 8}, got {new_value.bit_length()}"
            )
            self._value = new_value

        else:
            raise TypeError(f"Unsupported value type {type(new_value)} for the field value.")

    def __post_init__(self):
        """Perform sanity checks on the field properties."""
        self.value = 0

        if self.size < 0:
            raise ValueError("Field size must be a positive integer.")

        if self.offset < 0:
            raise ValueError("Field offset must be a positive integer.")

        # The last byte index that is occupied by this field.
        self.end = self.offset + self.size - 1

    def __hash__(self) -> int:
        """Hash functionality."""
        return hash((self.name, self.offset, self.size))


class PacketHeader:
    """An iterable collection of Field objects that forms the packet header."""

    def __init__(self, fields: list[Field]):
        """Initialize the header fields. Add more fields to it by means of ``add()``.

        Instantiate via:
        PacketHeader(
            [
                Field("field_name", 0, 4),
                Field("next_field_name", 4, 1),
                # ...
            ]
        )

        Args:
            fields (List[Field]): The list of fields to add initially.
        """
        self._fields: dict[FieldName, Field] = {}

        for field in fields:
            self.add_field(field)

    def copy_from_template(self, template: PacketHeader):
        """Copy the content of all matching fields from the template to this header.

        Args:
            template (PacketHeader): The header to copy from.
        """
        for field in template:
            if field.name in self:
                self[field.name] = field.value

    @property
    def size(self) -> int:
        """The total size of the header in bytes."""
        return sum(field.size for field in self)

    @property
    def is_continuous(self) -> bool:
        """Whether or not there are spaces in the header that are not defined."""
        fields_entries = self.as_list()

        return all(field.end + 1 == next_field.offset for field, next_field in zip(fields_entries, fields_entries[1:]))

    def _check_for_overlaps(self):
        """Check for overlapping fields.

        Raises:
            MemoryError: If overlapping fields are found.
        """
        fields_entries = self.as_list()

        # Check for overlapping fields, which are sorted by their offset.
        for field, next_field in zip(fields_entries, fields_entries[1:]):
            if not field.end <= next_field.offset:
                raise MemoryError("Fields {field.name} and {next_field.name} overlap.")

    def _sort_fields_by_offset(self):
        """Sorts the fields in this header by their offset."""
        self._fields = dict(sorted(self._fields.items(), key=lambda item: item[1].offset))

    def add(self, name: FieldName, offset: int, size: int):
        """Create and add a new field.

        Args:
            name (ValidFieldNames): The name of the field.
            offset (int): The field offset in number of bytes.
            size (int): The size of the field in bytes.
        """
        field = Field(name, offset, size)
        self.add_field(field)

    def add_field(self, field: Field):
        """Add a new field. Duplicates are ignored.

        Args:
            field (Field): The field to add.
        """
        if field not in self:
            self._fields[field.name] = field

            self._sort_fields_by_offset()
            self._check_for_overlaps()

    def as_bytes(self) -> bytes:
        """Get the full header as a bytes object."""
        buffer = bytearray()

        for field in self:
            int_to_bytes(field.value, buffer, field.offset, field.size)

        return bytes(buffer)

    def as_list(self) -> list[Field]:
        """The fields as a list.

        Returns:
            List[Field]: The list of fields.
        """
        return list(self._fields.values())

    def parse(self, data: bytes):
        """Parse a header and populate field values.

        Args:
            data (bytes): The data to parse.
        """
        if len(data) != self.size:
            raise ValueError(f"Input data needs to be exactly {self.size} bytes long!")

        for field in self:
            field.value = bytes_to_int(data, field.offset, field.size)

    @property
    def is_write_request(self) -> bool:
        """Whether this is a write request."""
        return self["operation"].value == OperationKey.WRITE_KEY.value

    @property
    def is_read_request(self) -> bool:
        """Whether this is a read request."""
        return self["operation"].value == OperationKey.READ_REQUEST_KEY.value

    @property
    def is_read_response(self) -> bool:
        """Whether this is a read response."""
        return self["operation"].value == OperationKey.READ_RESPONSE_KEY.value

    @property
    def is_safeload(self) -> bool:
        """Whether this is a software-safeload write request."""
        return self.is_write_request and self["safeload"].value == 1

    @property
    def carries_payload(self) -> bool:
        """Whether the corresponding packet carries a payload."""
        return self.is_write_request or self.is_read_response

    def __iter__(self) -> Iterator[Field]:
        """The iterator for fields."""
        yield from self._fields.values()

    def __setitem__(self, name: FieldName, value: int | bytes | bytearray):
        """Set a field value.

        Args:
            name (ValidFieldNames): Field name.
            value (int | bytes | bytearray): Field value.
        """
        assert name in self, f"Invalid field name {name}; valid names are '{self._fields.keys()}'"

        # FIXME: Type is correct, see https://github.com/python/mypy/issues/3004.
        # The value attribute has different types for setting and getting. During setting, there is a conversion step.
        self._fields[name].value = value  # type: ignore

    def __getitem__(self, name: FieldName) -> Field:
        """Get a field by its name.

        Args:
            name (ValidFieldNames): The name of the field.

        Returns:
            Field: The field, or None, if no field was found.

        Raises:
            IndexError: If the field does not exist.
        """
        assert name in self, f"Invalid field name {name}; valid names are '{self._fields.keys()}'"

        return self._fields[name]

    def __contains__(self, name: FieldName) -> bool:
        """Magic methods for using ``in``.

        Args:
            name (ValidFieldNames): The name to look for in the fields.

        Returns:
            bool: True, if a field with the given name exists.
        """
        return name in self._fields


class PacketHeaderGenerator(ABC):
    """Generic generator for packet headers."""

    @staticmethod
    @abstractmethod
    def _new_write_header() -> PacketHeader:
        """Generate a new header for a write packet."""

    @staticmethod
    @abstractmethod
    def _new_read_request_header() -> PacketHeader:
        """Generate a new header for a read request packet."""

    @staticmethod
    @abstractmethod
    def _new_read_response_header() -> PacketHeader:
        """Generate a new header for a read response packet."""

    def new_header_from_operation_byte(
        self, operation_byte: bytes, template: PacketHeader | None = None
    ) -> PacketHeader:
        """Generate a header from an operation byte.

        Args:
            operation_byte (bytes): The operation byte that determines the header composition.
            template (PacketHeader | None, optional): The template, if any, from which to copy fields that fit the
                new header. Defaults to None.

        Returns:
            PacketHeader: The new packet header.
        """
        assert len(operation_byte) == 1, "Operation byte must have a length of 1."

        operation_key = OperationKey(bytes_to_int(operation_byte, offset=0, length=1))

        header: PacketHeader
        if operation_key == OperationKey.READ_REQUEST_KEY:
            header = self._new_read_request_header()

        elif operation_key == OperationKey.READ_RESPONSE_KEY:
            header = self._new_read_response_header()

        elif operation_key == OperationKey.WRITE_KEY:
            header = self._new_write_header()

        else:
            raise ValueError(f"Unknown operation key {operation_key}.")

        if template is not None:
            header.copy_from_template(template)

        header["operation"] = operation_byte

        return header

    def new_header_from_operation_key(
        self, operation_key: OperationKey, template: PacketHeader | None = None
    ) -> PacketHeader:
        """Generate a header from an operation key.

        Args:
            operation_key (OperationKey): The operation key that determines the header composition.
            template (PacketHeader | None, optional): The template, if any, from which to copy fields that fit the
                new header. Defaults to None.

        Returns:
            PacketHeader: The new packet header.
        """
        return self.new_header_from_operation_byte(int8_to_bytes(operation_key.value), template=template)
